import numpy as np
import matplotlib.pyplot as plt
import argparse

def read_data(filename):
    """Reads a data file and returns it as a 2D numpy array."""
    with open(filename, 'r') as file:
        lines = file.readlines()

    size = int(lines[-2].split()[0]) + 1
    # Initialize empty lists to store rows
    data = np.zeros((size,size))
    # Read lines and parse the data
    for line in lines:
        if line.strip():  # If the line is not empty
            data[int(line.split()[0])][int(line.split()[1])] = float(line.split()[2])

    return data

def plot_data(data, output_file=None):
    """Plots the 2D data array using matplotlib."""
    plt.figure(figsize=(8, 6))
    plt.imshow(data, cmap='viridis', interpolation='nearest')
    plt.colorbar(label='Value')
    plt.title('2D Field Plot')
    plt.xlabel('X Coordinate')
    plt.ylabel('Y Coordinate')
    plt.grid(True)

    
    if output_file:
        plt.savefig(output_file, bbox_inches='tight')
        print(f"Plot saved to {output_file}")
    else:
        plt.show()


def main():
    # Specify the filename
    filename = 'u-exit.dat'  # Change this to your filename
    
    parser = argparse.ArgumentParser(description='Plot a 2D field from a dimepack data file.')
    parser.add_argument('filename', type=str, help='Path to the data file')
    parser.add_argument('--output', type=str, help='Path to the output file (e.g., plot.png)', default=None)

    args = parser.parse_args()
    
    data = read_data(args.filename)

    plot_data(data, output_file=args.output)
    
if __name__ == "__main__":
    main()
